/*************************************************************
 *                 bbomberman                                *
 *  (c) 2004 by Bernhard Eder (bb-generation)                *
 *  http://bbgen.net/bomberman for updates                   *
 *  Webmaster@bbgen.net for bug reports and other stuff      *
 * --------------------------------------------------------- *
 * This program is free software; you can modify it under    *
 * the terms of the GNU General Public License as published  *
 * by the Free Software Foundation, located in LICENSE       *
 *                                                           *
 *************************************************************/

#ifndef PROTO_H
#define PROTO_H

/* headers needed for defining the following classes */
#include <string>
#include <list>
#include <vector>


/*
 * BBomberman network protocol definition
 *****************************************
 * messages are sent as strings over the network;
 * each message is of the following format:
 *
 *   "MESSAGE_NAME[ ARG1[ ARG2[ ARG3[...]]]]\n"
 *
 */

/* network message numeric representation;
 * c: client -> server; s: server -> client
 */
enum bb_proto_msg
{
  BB_PROTO_MSG_DO_BBMOVE  =  1,  /* c: move bomberman
                                    arg: int bomberman, int direction */
  BB_PROTO_MSG_DO_BBBOMB  =  2,  /* c: set bomb
                                    arg: int bomberman                */

  BB_PROTO_MSG_EV_BBMOVE  = 33,  /* s: move bomberman
                                    arg: int bomberman, int direction */
  BB_PROTO_MSG_EV_BBBOMB  = 34,  /* s: set bomb
                                    arg: int bomberman                */
  BB_PROTO_MSG_EV_BBPUP   = 35,  /* s: bomberman got powerup
                                    arg: int bomberman, bb_dynamic_field powerup   */
  BB_PROTO_MSG_EV_BBKILL  = 36,  /* s: bomberman killed
                                    arg: int bomberman, int killer                 */

  BB_PROTO_MSG_EV_MAPFIELD  = 41,  /* s: map field changed
                                      arg: int x, int y, bb_map_field field_kind     */
  BB_PROTO_MSG_EV_DFCREATE  = 42,  /* s: dynamic field erased
                                      arg: int x, int y, bb_dynamic_field field_kind */
  BB_PROTO_MSG_EV_DFERASE   = 43,  /* s: dynamic field erased
                                      arg: int x, int y, bb_dynamic_field field_kind */
  BB_PROTO_MSG_EV_TICK      = 44,  /* s: advance to next tick (_should_ be (last + 1),
                                        but clients should not rely on this)
                                      arg: int tick                                  */
  BB_PROTO_MSG_EV_DELBOMB   = 45,  /* s: bomb object has been deleted in the server
                                    * arg: int bman_id[FIXME:, int bomb_index]
                                    */

  BB_PROTO_MSG_DENY_BBMOVE  = 49,  /* s: bomberman move denied
                                      arg: int bomberman              */
  BB_PROTO_MSG_DENY_BBBOMB  = 50,  /* s: bomberman bomb denied
                                      arg: int bomberman              */

  /* BB_PROTO_MSG_INIT_* messages shall only be used during client
   * initialization
   */
  BB_PROTO_MSG_INIT_MAPCOL = 65,  /* s: full map column state
                                     arg: int x, string with one char for each field:
                                       E: empty, B: box, F: fixed    */
  BB_PROTO_MSG_INIT_BMAN   = 66,  /* s: copy of a bomberman
                                   * arg: int id, str name, bool alive,
                                   *   int x, int y, bb_direction direction,
                                   *   int bombs_max, int expforce
                                   */
  BB_PROTO_MSG_INIT_BOMB   = 67,  /* s: copy of a bomb
                                   * arg: long long creation_tick, int id,
                                   *   int x, int y
                                   */
  BB_PROTO_MSG_INIT_DONE   = 68   /* s: initialization done; no further INIT_*
                                   *   messages will be sent
                                   * arg: (none)
                                   */
};


/* internally, messages are stored as cl_proto_msg */
class cl_proto_msg
{
  public:
    typedef  std::vector<std::string>  args_type;

    /* constructors, destructors */
    cl_proto_msg(const bb_proto_msg type);
    cl_proto_msg(const bb_proto_msg type, const cl_proto_msg& old_msg);
    cl_proto_msg(const bb_proto_msg type, const args_type& args);
    cl_proto_msg(const std::string& msg_str);

    /* public member functions */
    std::string&        arg(const size_t n);
    const std::string&  arg(const size_t n) const;
    const int           arg_to_int(const size_t n) const;
    args_type           args(void) const;
    void                args(args_type new_args);
    size_t              args_count(void) const;
    void                clear_args(void);
    void                decode(const std::string& msg_str);
    std::string         encode(void) const;
    std::string         escape(const std::string& str) const;
    void                push_arg(const std::string& arg);
    /* FIXME: use templates? */
    void                push_arg(const int arg);
    bb_proto_msg        type(void) const;
    void                type(const bb_proto_msg type);

  private:
    /* encapsulated data */
    bb_proto_msg        _M_type;
    args_type           _M_args;

    /* private member functions */
    void                _M_check_arg_index(const size_t n) const;
};

/* a queue for storing & retrieving cl_proto_msg objects */
class cl_proto_queue : private std::list<cl_proto_msg>
{
  public:
    bool          empty(void) const;
    std::string   poll_encode_msg(void);
    cl_proto_msg  poll_msg(void);
    void          queue_decode_msg(const std::string& msg_str);
    void          queue_msg(const cl_proto_msg& msg);
    size_t        size(void) const;
};

#endif
