/*************************************************************
 *                 bbomberman                                *
 *  (c) 2004 by Bernhard Eder (bb-generation)                *
 *  http://bbgen.net/bomberman for updates                   *
 *  Webmaster@bbgen.net for bug reports and other stuff      *
 * --------------------------------------------------------- *
 * This program is free software; you can modify it under    *
 * the terms of the GNU General Public License as published  *
 * by the Free Software Foundation, located in LICENSE       *
 *                                                           *
 *************************************************************/

#ifndef SERVER_H
#define SERVER_H

#include <string>
#include <sstream>
#include <list>
#include <vector>
#include <map>
#include <exception>
#include <stdexcept>

#include "cl_proto.h"
#include "ref_ptr.h"


enum bb_direction
{
  BB_DIRECTION_NONE  = -1,
  BB_DIRECTION_LEFT  =  0,
  BB_DIRECTION_RIGHT =  1,
  BB_DIRECTION_UP    =  2,
  BB_DIRECTION_DOWN  =  3
};

enum bb_map_field
{
  BB_MAP_FIELD_EMPTY        = 0,
  BB_MAP_FIELD_BOX          = 1,
  BB_MAP_FIELD_BOX_FIXED    = 2
};

enum bb_dynamic_field
{
  BB_DYN_FIELD_PUP_BOMBSMAX = 3,
  BB_DYN_FIELD_PUP_EXPFORCE = 4,
  BB_DYN_FIELD_EXPLOSION    = 5
};

enum bb_bomb_status
{
  BB_BOMB_STAT_TICKING1  = 0,
  BB_BOMB_STAT_TICKING2  = 1,
  BB_BOMB_STAT_EXPLODING = 2,
  BB_BOMB_STAT_VANISHED  = 3
};

struct bb_map_field_xy
{
  int x, y;

  /* constructor */
  bb_map_field_xy(const int x, const int y) : x(x), y(y) { };

  /* lesser operator overload */
  bool  operator<(const bb_map_field_xy& other) const
  {
    /* x coordinate lesser? */
    if (x < other.x)
      return true;

    /* x coordinate equal && y coordinate lesser? */
    if ((x == other.x) && (y < other.y))
      return true;

    /* other is not lesser than this */
    return false;
  }

  /* equal operator overload */
  bool  operator==(const bb_map_field_xy& other) const
  {
    /* are other coordinates equal? */
    if ((x == other.x) && (y == other.y))
      return true;

    /* other is not equal to this */
    return false;
  }
  bool  operator!=(const bb_map_field_xy& other) const { return !operator==(other); }
};

/* cl_client instance kind (for polymorphism) */
enum bb_client_kind
{
  BB_CLIENT_LOCAL  = 0,  /* basic (local) client                   */
  BB_CLIENT_REMOTE = 1,  /* remote client used in cl_server_net    */
  /* FIXME: won't be used.. */
  BB_CLIENT_SERVER = 2   /* pseudo client used in cl_server_remote */
};


/* prototype all classes so they can be used by each other */
class cl_bomb;
class cl_bomberman;
class cl_client;
class cl_server;
class cl_server_local;


class cl_bomb
{
  public:
    cl_bomberman&   owner;
    const int       x, y;

    cl_bomb(cl_bomberman& bomb_owner);

    long long       creation_tick() const;
    void            explode();
    bool            explode_field(const int x, const int y, const bb_direction& direction);
    bool            explode_to(const bb_direction& direction);
    bb_bomb_status  status() const;

  protected:
    cl_bomb(cl_bomberman& bomb_owner, const int x, const int y, const long long creation_tick);

  private:
    long long       _M_creation_tick;

    void            creation_tick(const long long new_creation_tick);
};

class cl_bomberman
{
  friend class cl_bomb;
  friend class cl_server;
  friend class cl_server_local;
  friend class cl_server_remote;

  public:
    typedef  ref_ptr<cl_bomb>       bomb_ptr;
    typedef  std::vector<bomb_ptr>  bombs_type;

    cl_server&           server;
    const int            id;
    const std::string    name;
    bool                 alive;  // True if the bomberman is alive
    int                  x, y;
    bb_direction         direction;
    bombs_type           bombs;
    int                  bombs_max;
    int                  expforce;
    cl_client*           client;
    /* FIXME: really useful after cl_game now parsing its client's to_client
     *   queue..?
     */
    const cl_bomberman*  killed_by;

    cl_bomberman(cl_server& server, const int id, const std::string& name, cl_client* client = NULL);

    void           move(const bb_direction& to) const;
    void           place_bomb() const;

  protected:
    cl_bomberman(cl_server& server, const int id, const std::string& name, const bool alive, const int x, const int y, const bb_direction direction, const int bombs_max, const int expforce, cl_client* client = NULL);

    void  do_kill(const cl_bomberman* killer);
    bool  do_move(const bb_direction& to);
    bool  do_place_bomb();
    void  do_powerup(const bb_dynamic_field powerup);
};

class cl_client
{
  friend class cl_bomberman;
  friend class cl_server;
  friend class cl_server_local;

  public:
    /* FIXME: language; use C++ language features to determine real class name of polymorphic class instance */
    const bb_client_kind  kind;
    cl_server&            server;

    cl_client(cl_server& server);
    virtual ~cl_client() { }

    /* FIXME: language; this should be encapsulated...
     *   it's not as easy as protecting it, as e.g. cl_game needs to read the
     *   to_client queue
     */
    cl_proto_queue  to_client;
    cl_proto_queue  to_server;

    virtual void    detach();

  protected:
    cl_client(const bb_client_kind kind, cl_server& server);
};

class cl_server
{
  public:
    typedef  ref_ptr<cl_client>                       client_ptr;
    typedef  std::vector<client_ptr>                  clients_type;
    typedef  ref_ptr<cl_bomberman>                    bomberman_ptr;
    typedef  std::vector<bomberman_ptr>               bombermans_type;
    typedef  std::vector<std::vector<bb_map_field> >  map_type;
    typedef  std::multimap<bb_map_field_xy, bb_dynamic_field>
                                                      dynamic_fields_type;

    clients_type     clients;
    bombermans_type  bombermans;

    virtual ~cl_server() { }

    cl_bomberman&        bomberman(const int id) const;
    dynamic_fields_type  dynamic_fields() const;
    dynamic_fields_type  dynamic_fields(const int x, const int y) const;
    virtual void         explode_field(const int x, const int y, const cl_bomberman* causer) = 0;
    virtual bool         is_delaying() const = 0;
    virtual bool         is_initialized() const = 0;
    bool                 is_field_passable(const int x, const int y) const;
    bb_map_field         map(const int x, const int y) const;
    int                  map_fields_x() const;
    int                  map_fields_y() const;
    virtual void         tick() = 0;
    long long            tick_count() const;
    
  protected:
    typedef  dynamic_fields_type::iterator        _dyn_fs_iterator;
    typedef  dynamic_fields_type::const_iterator  _dyn_fs_const_iterator;

    dynamic_fields_type  _M_dynamic_fields;
    map_type             _M_map;
    int                  _M_map_fields_x, _M_map_fields_y;
    long long            _M_ticks;

    virtual void         _M_create_dynamic_field(const int x, const int y, const bb_dynamic_field field_kind);
    virtual void         _M_destroy_dynamic_field(const int x, const int y, const bb_dynamic_field field_kind);
    std::string          _M_encode_map_column(const int x) const;
    void                 _M_decode_map_column(const std::string& in);
};

class cl_server_local : public cl_server
{
  public:
    typedef  std::vector<std::string>   bomberman_names_type;

    cl_server_local(const int map_fields_x, const int map_fields_y, const bomberman_names_type& bomberman_names);
    ~cl_server_local();

    void  explode_field(const int x, const int y, const cl_bomberman* causer);
    bool  is_delaying() const { return false; }
    bool  is_initialized() const { return true; }
    void  tick();

  private:
    void  _M_bomberman_powerup(cl_bomberman& bomberman);
    void  _M_build_map();
    void  _M_create_dynamic_field(const int x, const int y, const bb_dynamic_field field_kind);
    void  _M_destroy_dynamic_field(const dynamic_fields_type::iterator& field);
    void  _M_destroy_dynamic_field(const int x, const int y, const bb_dynamic_field field_kind);
    void  _M_destroy_dynamic_fields(const bb_dynamic_field field_kind);
    void  _M_process_bombs();
    void  _M_process_client_msgs();
    void  _M_queue_for_clients(const cl_proto_msg& msg);
    void  _M_set_map_field(const int x, const int y, const bb_map_field field_kind);
};

#endif
