/*************************************************************
 *                 bbomberman                                *
 *  (c) 2004 by Bernhard Eder (bb-generation)                *
 *  http://bbgen.net/bomberman for updates                   *
 *  Webmaster@bbgen.net for bug reports and other stuff      *
 * --------------------------------------------------------- *
 * This program is free software; you can modify it under    *
 * the terms of the GNU General Public License as published  *
 * by the Free Software Foundation, located in LICENSE       *
 *                                                           *
 *************************************************************/
 
#include "cl_server_net.h"

/* used in implementation only */
#include <exception>
#include <stdexcept>

/* FIXME: debug */
#include <iostream>


/*
 * cl_client_net implementation
 */

/* constructor */
cl_client_net::cl_client_net(cl_server_net& server, TCPsocket socket) :
  cl_client(BB_CLIENT_REMOTE, server), server(server), _M_socket(socket), _M_connected(true)
{
  /* initialize network buffer */
  for (int i = 0; i < BB_NET_BUF_LEN; i++)
    _M_buffer[i] = '\0';

  _M_buffer_insert_ptr = _M_buffer;

  /* add client socket to server's SDL_net socket set */
  SDLNet_TCP_AddSocket(server._M_socket_set, _M_socket);
}

/* destructor */
cl_client_net::~cl_client_net()
{
  /* still connected to real client? */
  if (_M_connected)
  {
    /* close network connection to real client */
    SDLNet_TCP_Close(_M_socket);
    _M_connected = false;
  }

  /* remove client socket from server's SDL_net socket set */
  SDLNet_TCP_DelSocket(server._M_socket_set, _M_socket);
}

/* FIXME: maybe don't disconnect? so client would get a chance to process a
 *   connection-aborted / error message? maybe mark for immediate disconnection
 *   after this message has been passed?
 */
/* detach client from server */
void cl_client_net::detach()
{
  std::cerr << "* cl_client_net::detach() -- detaching networked client" << std::endl;

  /* detach from local server objects (inherited from cl_client) */
  cl_client::detach();

  /* still connected? */
  if (_M_connected)
  {
    /* close network connection to real client */
    SDLNet_TCP_Close(_M_socket);
    _M_connected = false;
  }
}

/* pass queued protocol messages on to real client via network */
void cl_client_net::pass_proto_msgs()
{
  while (!to_client.empty())
  {
    cl_proto_msg  msg    = to_client.poll_msg();
    std::string   data   = msg.encode();
    data += '\n';

    /* FIXME: cast; explicit typecast from (const void*) to (void*); sadly, SDL_net API isn't const-ified. :'(
     *   assuming SDLNet_TCP_Send() won't modify its send buffer
     */
    int  length = data.length();
    int  sent = SDLNet_TCP_Send(_M_socket, (void*) data.data(), length);

    if (sent < length)
      throw std::runtime_error("cl_client_net::pass_proto_msgs(): unknown network error");
  }
}

/* process net events for this client */
bool cl_client_net::process_net_events()
{
  /* data received on client's socket? */
  if (!SDLNet_SocketReady(_M_socket))
  {
    /* indicate no net events for this client available */
    return false;
  }

  /* process incoming data */
  do
  {
    /* read BB_NET_BUF_LEN bytes from network */
    int size = SDLNet_TCP_Recv(_M_socket, _M_buffer_insert_ptr, BB_NET_BUF_LEN - (_M_buffer_insert_ptr - _M_buffer));
    char*  buffer_begin_ptr = _M_buffer;
    char*  buffer_end_ptr   = _M_buffer_insert_ptr + size;

    /* check for receive error */
    if (size <= 0)
      throw std::runtime_error("cl_client_net::process_net_events(): unknown network error");

    /* process received data */
    for (char* buffer_ptr = _M_buffer_insert_ptr;
         buffer_ptr < buffer_end_ptr;
         buffer_ptr++)
    {
      /* end-of-line encountered? */
      if (*buffer_ptr == '\n')
      {
        /* convert line to NUL-terminated C string */
        *buffer_ptr = '\0';

        /* convert C string to std::string & enqueue in message queue */
        //to_server.queue_msg(cl_proto_msg(std::string(buffer_begin_ptr)));
        std::string  line(buffer_begin_ptr);
        std::cerr << "* cl_client_net::process_net_events(): received: \"" <<
          line << "\"" << std::endl;
        to_server.queue_msg(cl_proto_msg(line));

        /* begin new line after NUL C string terminator */
        buffer_begin_ptr = buffer_ptr + 1;
      }
    }

    /* at least one complete line found in buffer? */
    if (buffer_begin_ptr != _M_buffer)
    {
      /* move trailing chars to begin of buffer */
      for (char* buffer_ptr = buffer_begin_ptr;
           buffer_ptr < buffer_end_ptr;
           buffer_ptr++)
      {
        _M_buffer[buffer_ptr - buffer_begin_ptr] = *buffer_ptr;
      }
    }
    else
    {
      /* check for line length exceeded */
      if ((buffer_end_ptr - buffer_begin_ptr) == BB_NET_BUF_LEN)
        throw std::runtime_error("cl_client_net::process_net_events(): maximum line length exceeded");
    }

    /* calculate new insert pointer */
    _M_buffer_insert_ptr = _M_buffer + (buffer_end_ptr - buffer_begin_ptr);
  }
  while (SDLNet_SocketReady(_M_socket));

  /* indicate net events for this client processed */
  return true;
}


/*
 * cl_server_net implementation
 */

/* constructor */
cl_server_net::cl_server_net(const int map_fields_x, const int map_fields_y, const bomberman_names_type& bomberman_names, const Uint16 port) :
  /* inherit basic constructor from cl_server_local */
  cl_server_local(map_fields_x, map_fields_y, bomberman_names)
{
  /* initialize network API */
  if (SDLNet_Init())
    throw std::runtime_error("cl_server_net::cl_server_net(): can't initialize SDL_net");

  /* listen on server socket */
  IPaddress  ip_addr;
  ip_addr.host = INADDR_ANY;
  SDLNet_Write16(port, &ip_addr.port);
  _M_server_socket = SDLNet_TCP_Open(&ip_addr);

  if (_M_server_socket == NULL)
    throw std::runtime_error("cl_server_net::cl_server_net(): can't listen on server socket");

  /* initialize socket set (for polling data on all sockets without blocking) */
  /* FIXME: const */
  _M_socket_set = SDLNet_AllocSocketSet(5);

  if (_M_socket_set == NULL)
    throw std::runtime_error("cl_server_net::cl_server_net(): can't allocate socket set");

  SDLNet_TCP_AddSocket(_M_socket_set, _M_server_socket);
}

/* destructor */
cl_server_net::~cl_server_net()
{
  /* free socket set */
  SDLNet_FreeSocketSet(_M_socket_set);

  /* stop listening on server socket */
  SDLNet_TCP_Close(_M_server_socket);

  /* cleanup network API */
  SDLNet_Quit();
}

void cl_server_net::tick()
{
  int socket_count;

  /* poll network sockets */
  switch (socket_count = SDLNet_CheckSockets(_M_socket_set, 0))
  {
    case -1:
      /* error in library function */
      throw std::runtime_error("cl_server_net::tick(): unknown network error");

    case 0:
      /* no network activity */
      break;

    default:
      /* process network events */
      _M_process_net_events(socket_count);
  }

  /* invoke inherited tick() */
  cl_server_local::tick();

  /* pass protocol messages on to clients */
  _M_pass_proto_msgs();
}

/* accept new client */
void cl_server_net::_M_accept_client()
{
  /* accept client network connection to server socket */
  TCPsocket  client_socket = SDLNet_TCP_Accept(_M_server_socket);

  if (client_socket == NULL)
    throw std::runtime_error("cl_server_net::_M_accept_client(): can't accept incoming connection");

  /* create new client object for communication with client */
  cl_client_net*  client = new cl_client_net(*this, client_socket);
  clients.push_back(client_ptr(client));

  /* FIXME: also send bomberman names (etc.?) */
  /* queue initialization messages for client */
  _M_send_map(client->to_client);
  _M_send_dynamic_fields(client->to_client);
  _M_send_bombermans(client->to_client);
  _M_send_bombs(client->to_client);
  
  /* indicate initialization done */
  client->to_client.queue_msg(cl_proto_msg(BB_PROTO_MSG_INIT_DONE));
}

/* pass protocol messages on to clients */
void cl_server_net::_M_pass_proto_msgs()
{
  /* iterate through all clients */
  for (clients_type::iterator  client = clients.begin();
       client < clients.end(); client++)
  {
    /* skip non-remote clients */
    if ((*client)->kind != BB_CLIENT_REMOTE)
      continue;

    /* access remote client specialization interface */
    cl_client_net*  client_net = (cl_client_net*) client->get();

    try
    {
      /* let client object pass its messages on to real client via network */
      client_net->pass_proto_msgs();
    }
    catch (std::runtime_error err)
    {
      /* assume network error -- detach client */
      std::cerr << "* cl_server_net::_M_pass_proto_msgs(): error while sending queued messages to networked client" << std::endl;
      client_net->detach();
      clients.erase(client);
    }
  }
}

/* process network events */
void cl_server_net::_M_process_net_events(int socket_count)
{
  /* incoming connection? */
  if (SDLNet_SocketReady(_M_server_socket))
  {
    /* accept new clients */
    do
    {
      _M_accept_client();
    }
    while (SDLNet_SocketReady(_M_server_socket));

    socket_count--;
  }

  if (socket_count < 1)
    return;

  /* iterate through clients */
  for (clients_type::iterator  client = clients.begin();
       (socket_count > 0) && (client < clients.end());
       client++)
  {
    /* skip non-remote clients */
    if ((*client)->kind != BB_CLIENT_REMOTE)
      continue;

    /* access remote client specialization interface */
    cl_client_net*  client_net = (cl_client_net*) client->get();

    /* catch network errors */
    try
    {
      /* try network event processing for this client */
      if (client_net->process_net_events())
      {
        /* client succeeded in processing net events,
         * so there's one possible socket less for other net events
         */
        socket_count--;
      }
    }
    /* FIXME: except; use own exception class */
    catch (std::runtime_error err)
    {
      /* assume network error -- detach client */
      std::cerr << "* cl_server_net::_M_process_net_events(): error while processing messages from networked client" << std::endl;
      client_net->detach();
      clients.erase(client);
    }
  }
}

/* send copy of all bombermans to client */
void cl_server_net::_M_send_bombermans(cl_proto_queue& queue) const
{
  for (bombermans_type::const_iterator  bman_ptr = bombermans.begin();
       bman_ptr < bombermans.end(); bman_ptr++)
  {
    cl_bomberman&  bman = **bman_ptr;
    cl_proto_msg   msg(BB_PROTO_MSG_INIT_BMAN);

    msg.push_arg(bman.id);
    msg.push_arg(bman.name);
    msg.push_arg(bman.alive);
    msg.push_arg(bman.x);
    msg.push_arg(bman.y);
    msg.push_arg(bman.direction);
    msg.push_arg(bman.bombs_max);
    msg.push_arg(bman.expforce);

    queue.queue_msg(msg);
  }
}

/* send copy of all bombs to client */
void cl_server_net::_M_send_bombs(cl_proto_queue& queue) const
{
  for (bombermans_type::const_iterator  bman_ptr = bombermans.begin();
       bman_ptr < bombermans.end(); bman_ptr++)
  {
    cl_bomberman&  bman = **bman_ptr;

    for (cl_bomberman::bombs_type::const_iterator  bomb_ptr = bman.bombs.begin();
         bomb_ptr < bman.bombs.end(); bomb_ptr++)
    {
      cl_bomb&      bomb = **bomb_ptr;
      cl_proto_msg  msg(BB_PROTO_MSG_INIT_BOMB);

      msg.push_arg(bomb.creation_tick());
      msg.push_arg(bomb.owner.id);
      msg.push_arg(bomb.x);
      msg.push_arg(bomb.y);

      queue.queue_msg(msg);
    }
  }
}

/* send copy of all currently existing dynamic fields to client */
void cl_server_net::_M_send_dynamic_fields(cl_proto_queue& queue) const
{
  for (dynamic_fields_type::const_iterator  field = _M_dynamic_fields.begin();
       field != _M_dynamic_fields.end(); field++)
  {
    cl_proto_msg  msg(BB_PROTO_MSG_EV_DFCREATE);
    msg.push_arg(field->first.x);
    msg.push_arg(field->first.y);
    msg.push_arg(field->second);
    queue.queue_msg(msg);
  }
}

/* send copy of map to client */
void cl_server_net::_M_send_map(cl_proto_queue& queue) const
{
  for (int x = 0; x < _M_map_fields_x; x++)
  {
    cl_proto_msg  msg(BB_PROTO_MSG_INIT_MAPCOL);
    msg.push_arg(_M_encode_map_column(x));
    queue.queue_msg(msg);
  }
}
