/*************************************************************
 *                 bbomberman                                *
 *  (c) 2004 by Bernhard Eder (bb-generation)                *
 *  http://bbgen.net/bomberman for updates                   *
 *  Webmaster@bbgen.net for bug reports and other stuff      *
 * --------------------------------------------------------- *
 * This program is free software; you can modify it under    *
 * the terms of the GNU General Public License as published  *
 * by the Free Software Foundation, located in LICENSE       *
 *                                                           *
 *************************************************************/

#ifndef REF_PTR_H
#define REF_PTR_H
 
/* needed for definition of NULL... ._. */
#include <ios>


/*
 * A pointer with reference counting.
 *
 * When the last ref_ptr object referencing the object which is
 * pointed to is destroyed, the object which is pointed to will
 * be deleted.
 */
template<typename _T>
class ref_ptr
{
  public:
    /* constructor (from pointer) */
    ref_ptr(_T* ptr) :
      _M_ptr(ptr)
    {
      /* create new reference counter & initialize it to 1 */
      _M_ref_count = new int(1);
    }

    /* constructor (from other ref_ptr) */
    ref_ptr(const ref_ptr& other_ref_ptr)
    {
      _M_copy(other_ref_ptr);
    }

    /* destructor */
    ~ref_ptr()
    {
      /* destruct contained data unless already done (via e.g. release()) */
      if (_M_ref_count)
        _M_destruct();
    }

    /* operator overloads */
    ref_ptr<_T>&  operator=(const ref_ptr& other_ref_ptr)
    {
      _M_destruct();
      _M_copy(other_ref_ptr);
      return *this;
    }
    _T&  operator*()  const { return *_M_ptr; }
    _T*  operator->() const { return  _M_ptr; }

    /* access raw pointer */
    _T*  get() const { return _M_ptr; }

    /* retrieve raw pointer & release ref_ptr from its deletion duties */
    _T*  release()
    {
      _T*  tmp_ptr = _M_ptr;
      _M_destruct();
      return tmp_ptr;
    }

  protected:
    _T*   _M_ptr;
    int*  _M_ref_count;

  private:
    void  _M_copy(const ref_ptr& other_ref_ptr)
    {
      _M_ptr       = other_ref_ptr._M_ptr;
      _M_ref_count = other_ref_ptr._M_ref_count;

      /* increase shared reference counter */
      ++(*_M_ref_count);
    }

    void  _M_destruct()
    {
      /* decrease shared reference counter */
      --(*_M_ref_count);

      /* destroy object we point to if necessary */
      if (*_M_ref_count < 1)
      {
        delete _M_ptr;
        delete _M_ref_count;
      }

      /* invalidate internal pointers */
      _M_ptr       = NULL;
      _M_ref_count = NULL;
    }
};

#endif
